//========================================================================
// SCIDYNE Corporation   649 School Street   Pembroke, MA 02359-3649 USA
// Tel: (781) 293-3059   Fax: (781) 293-4034   URL: www.scidyne.com
//========================================================================
/// NOTICE:
//    This work is licensed under a Creative Commons Attribution-ShareAlike 3.0 Unported License. 
//    http://creativecommons.org/licenses/by-sa/3.0/us/
//
//========================================================================
// Project  :  XMEM+                                     
// File     :  sram_insight.ino                                             
// Revision :  1.00                                               
// Date     :  12-30-15 
// Author:  :  Mark Durgin
// Target   :  Arduino / Genuino Mega 2560, Arduino Mega ADK, Uno
// Compiler :  Arduino IDE V1.6.6                                                
//========================================================================
// Description and usage: 
//  This program uses the IDE Serial Monitor to display the addresses and
//  relative sizes of the four Arduino SRAM memory segments:
//
//    .data  -- Initialized data
//    .bss   -- Un-Initialized data
//    heap   -- Dynamic memory, managed by program
//    stack  -- Temporary function variables, managed automatically  
// 
//  As released, this programs intended target is an Arduino MEGA2560 with
//  a XMEM+ board installed. The heap segment is relocated from its default
//  internal SRAM location to instead use the XMEM+ SRAM (bank-0). To have
//  the heap remain in internal SRAM or to use this program with another Arduino
//  device, such as an UNO, comment-out the out the conditional compilation
//  directive #define XMEM_PLUS.   
//
//  Web References:
//	http://www.nongnu.org/avr-libc/user-manual/malloc.html
//
//  http://andybrown.me.uk/2011/01/01/debugging-avr-dynamic-memory-allocation/
//
//========================================================================
// Issues, Concerns and "THINGS TO DO":
//  
//========================================================================
// History:
//  12-30-15  R0.00   - Project Begins
//========================================================================

// Comment-out to have heap reside in internal SRAM or to use this program on an UNO
#define XMEM_PLUS 

// **** Reference important system variables and constants
extern char *__data_start;          // Starting address of .data segment ( Initialized variables )
extern char *__data_end;            // Ending address of .data segment ( Initialized variables )

extern char *__bss_start;           // Starting address of .bss segment ( un-initialized variables )
extern char *__bss_end;             // Ending address of .bss segment ( un-initialized variables )

extern char *__heap_start;          // System default address of where heap starts.  Value provided by linker
extern char *__heap_end;            // System default address of where heap end.  Value provided by linker
extern char *__malloc_heap_start;   // Variable that sets starting address of heap.  On reset this gets loaded with __heap_start
extern char *__malloc_heap_end;     // Variable that sets ending address of heap. On reset this gets loaded with __heap_end
//extern char *__brkval;              // Not used in this example.   A variable that shows the highest point in RAM occupied by the heap.
//extern char *__flp;                 // Not used in this example.   heap Free-List-Pointer. 


//************************************************************************
void setup()
{
  
  /****************************************************************************************
    Program initialization
  *****************************************************************************************/   
  Serial.begin(9600);  // Setup communications to serial monitor

  Serial.println(F("SCIDYNE:  sram_insight.ino  V1.0"));
  Serial.println(F("Display Arduino SRAM Segments"));
  Serial.println(F("Note: All values shown in hexadecimal"));
  Serial.println();

  #ifdef XMEM_PLUS
    // Activate external memory and relocate heap to XMEM+ SRAM. Bank-0 automatically selected at reset by XMEM+ logic.
    XMCRB  =  0x00;                         // All of all PORTC pins act as upper address lines, A[15:8], disable Bus keeper function 
	  XMCRA  =  0x80;                         // Set SRE bit, Enable xmem+, No wait states.
    __malloc_heap_start = (char *) 0x2200;  // Defines the start address of SRAM on the XMEM+ 
    __malloc_heap_end   = (char *) 0xA1FF;  // Defines the end address of SRAM on the XMEM+
  #endif

 
  /*******************************************************************
    Display internal SRAM parameters
  ********************************************************************/
  Serial.println(F("**** Internal SRAM parameters ****"));
  Serial.print(F("RAMSTART               : ")); Serial.println(RAMSTART, HEX);
  Serial.print(F("RAMEND                 : ")); Serial.println(RAMEND, HEX);
  Serial.print(F("Internal SRAM Size     : ")); Serial.println((RAMEND - RAMSTART)+1, HEX);
  Serial.println();

  
  /*******************************************************************
    Display system .bss segment parameters
  ********************************************************************/ 
  // Display system .data segment parameters
  Serial.println(F("**** .data segment parameters ****"));
  Serial.print(F("__data_start address   : ")); Serial.println((unsigned int) &__data_start, HEX);
  Serial.print(F("__data_end address     : ")); Serial.println((unsigned int) &__data_end, HEX);
  Serial.print(F(".data segment size     : ")); Serial.println((((unsigned int) &__data_end - (unsigned int) &__data_start)+1), HEX);
  Serial.println();


  /*******************************************************************
    Display system .bss segment parameters
  ********************************************************************/
  Serial.println(F("**** .bss segment parameters ****"));
  Serial.print(F("__bss_start address    : ")); Serial.println((unsigned int) &__bss_start, HEX);
  Serial.print(F("__bss_end  address     : ")); Serial.println((unsigned int) &__bss_end, HEX);
  Serial.print(F(".bss segment size      : ")); Serial.println((((unsigned int) &__bss_end - (unsigned int) &__bss_start)+1), HEX);
  Serial.println();


  /*******************************************************************
    Display heap segment parameters
  ********************************************************************/
  Serial.println(F("**** heap segment parameters ****"));
  Serial.print(F("__heap_start (default) : ")); Serial.println((unsigned int) &__heap_start, HEX);
  Serial.print(F("__heap_end (default)   : ")); Serial.println((unsigned int) &__heap_end, HEX);
  Serial.print(F("__malloc_heap_start    : ")); Serial.println((unsigned int) __malloc_heap_start, HEX); 
  Serial.print(F("__malloc_heap_end      : ")); Serial.println((unsigned int) __malloc_heap_end, HEX);   
  
  Serial.print(F("heap size (maximum)    : "));
  // When __malloc_heap_end is 0  any dynamic memory calls assume the heap is located internally and below the stack
  if ( !__malloc_heap_end )  // Determine where heap is located
  {
    Serial.print(((unsigned int) RAMEND - (unsigned int) __malloc_heap_start)+1, HEX);  Serial.println(F(" [ Using internal SRAM ]"));
  }
  else
  {
    Serial.print((unsigned int)(__malloc_heap_end - __malloc_heap_start)+1, HEX);  Serial.println(F(" [ Using external XMEM+ SRAM ]"));
  }
   
  Serial.println();


  /*******************************************************************
    Display system stack segment parameters
  ********************************************************************/
  Serial.println(F("**** stack parameters ****"));
  Serial.print(F("Top of stack (RAMEND)  : ")); Serial.println((unsigned int) RAMEND, HEX);
  Serial.println();

}



/*******************************************************************
    Main loop. Not used.
 ********************************************************************/
void loop()
{
     // This is the main loop.
     //  Do nothing, forever
}



