/*********************************************************************************
 * SCIDYNE Corporation
 * XMEM+ Memory Array Demonstration Program
 * File: xmem_bare_bones_array.ino
 * 11-24-2020   V1.0
 * Author: Mark Durgin
 * 
 * This example shows how to create and use a large data array in XMEM+ memory 
 * WITHOUT the need to relocate the heap, use dynamic memory, bank switching, 
 * or rely on special libraries. 
 * 
 * An array of 16k 2-byte integers will be demonstrated. Any data type is
 * supported provided the underlying 32768 byte limit of BANK-0 is not exceeded. 
 * 
 * The program results are displayed on the IDE Serial Monitor.
 *********************************************************************************
 */

/*
 * Define important parameters
 */
#define XMEM_START   ( (void *) 0x2200 )   // Start of XMEM+ External SRAM memory space
#define arraySize    16384                 // An Arduino int variable requires 2 bytes, BANK-0 has 32768 bytes available.



/************************************************************************************************
 * The setup code executes once after a reset. 
 * When completed program execution passes to loop()
 */
void setup() {
 
  /* 
   * Setup for using XMEM+ External Memory and Expansion Bus.  
   * Note: The XMEM+ Bank Select Register (BSEL) defaults to BANK-0 (0) at reset. 
   */
  XMCRB  =  0x00;   // All of all PORTC pins act as upper address lines, A[15:8], disable Bus keeper function 
  XMCRA  =  0x80;   // Set SRE bit, Enable xmem+ memory, No Wait States.

  /*
   * Setup serial communication to the IDE Monitor
   */
  Serial.begin(9600);
  Serial.println(F("SCIDYNE XMEM+ - A Simple Bare-Bones Array Example"));
}


/************************************************************************************************
 * After setup runs once execution continues here
 * Put your main code in loop() which runs repeatedly
 */
void loop() {

  /*
   * Create a pointer to an array of data type int and initialize it to the beginning of XMEM+ memory. 
   * An Arduino int variable requires 2-bytes for storage and BANK-0 has 32768 bytes available. 
   * Therefore, the maximum int array size is 32768 / 2 = 16384 int locations (0 through 16383)
   */
  int *  bigIntArray  = XMEM_START; 


  /*
   * Use the array normally. 
   * For this example, a different int value is written to each array location.
   * We'll make it the current array index plus the number 12345
   */
  
  for (int i = 0; i < arraySize; i++) {
    bigIntArray[i] =  i + 12345;
  }
  
  /*
   * See if the correct value got stored. Any location can be examined
   * but let’s arbitrarily choose location 14567.
   * The stored value should be:  14567 + 12345 = 26912
   */
  Serial.print(F("The XMEM+ location contains: "));
  Serial.println(bigIntArray[14567]);

  while(1);  // Loop here forever

}
