/*********************************************************************************
 * SCIDYNE Corporation
 * XMEM+ Memory Demonstration Program
 * 3-26-2020   V1.0
 * File: xmem_large_array_across_multiple_banks.ino
 * Mark Durgin
 * 
 * This example shows how to create a large data array spanning multiple XMEM+ memory banks
 * but WITHOUT the need to relocate the heap, use dynamic memory routines, or special libraries. 
 * 
 * An array of floating point numbers will be demostrated.  All data types are supported 
 * provided the underlying 32768 bytes-per-bank limit is not exceeded.
 * 
 * Results are display on IDE Serial Monitor
 * 
 ************************************************************************************
 */

/*
 * Declare prototypes
 */
void  writeBigArray(unsigned long int index, float data);  // Write one float number into array at index location
float readBigArray(unsigned long int index);               // Returns one float number read from index location


/*
 * Define important parameters
 */
#define XMEM_START      ( (void *) 0x2200 )                  // Start of XMEM+ External SRAM memory space
#define BYTES_PER_BANK  32768                                // MAXIMUM number of bytes in one XMEM+ memory bank
#define ARRAYSIZE       (BYTES_PER_BANK / sizeof(float))     // Pre-Calculate number of locations a bank can hold based on data type
                                                             //   floats use 4 bytes each
                                                          
/*
 * Declare global variables
 */
unsigned long int arrayIndex;
unsigned char * BSEL  = (unsigned char *) 0xFF01;      // Pointer to Bank select register BSEL[3:0]

float *  bigFloatArray = XMEM_START;  //Create a pointer to an array of data type float and initalize it to the beginning of XMEM+ memory.


/*
 * The setup code executes once after a reset. 
 * When completed program execution passes to loop()
 */
void setup() {
 
  /* 
   * Setup for using External Memeory and Expansion Bus.  
   * Note: XMEM+ Bank Select Register (BSEL) defaults to BANK-0 at reset. 
   */
  XMCRB  =  0x00;   // All of all PORTC pins act as upper address lines, A[15:8], disable ATMEGA2560 Bus keeper function 
  XMCRA  =  0x80;   // Set SRE bit, Enable xmem+ memory, No Wait States.

  /*
   * Setup serial communication to IDE Monitor
   */
  Serial.begin(9600);
  Serial.println(F("SCIDYNE XMEM+ - Demonstrate a Large Array across multiple banks"));
  
}


/*
 * After setup runs once execution continues here.
 */
void loop() {

  /*
   * Local variables 
   */
  unsigned long int i;
  float returned_float_value;

 
  /*
   * Use the array normally.  
   * This demonstration writes to each location with a different float value (1 float occupies 4-bytes). 
   * For easier debugging, the number stored at a location will be Pi multipled by the location index.
   * Note: 130000 locations x 4-bytes each requires 520,000 bytes of SRAM storage!
   */
  Serial.println(F("Loading Array ..."));
  for (i = 0; i < 130000; i++) {
    writeBigArray ( i, (3.1415927 * i));
  }
  
  /*
   * See if the correct value got stored. Any location can be examined
   * but lets arbitraily choose location 98773.
   * The stored value should then be:  98773 * 3.1415927 = 310304.53
   */

    i = 98773;  // Pick a location
   
    Serial.print(F(" Expected: "));             // Calculate and display what should be stored
    Serial.println( i * 3.1415927 );
  
    returned_float_value = readBigArray(i);     // Read and dsiplay the stored value
    Serial.print(F(" Returned: "));             
    Serial.println(returned_float_value);

    Serial.println();                           // Show details about the bank and offset within the bank
    Serial.print("Location: ");
    Serial.println(i);
    Serial.print("    Bank: ");
    Serial.println(i / ARRAYSIZE);
    Serial.print("  Offset: ");
    Serial.println(i % ARRAYSIZE);

 
    while(1);  // Loop here forever

}



/* 
 * Write one float number into array at index location
 */
void  writeBigArray(unsigned long int index, float data) 
{
    *BSEL = index / ARRAYSIZE;    // Set the Bank Select Register
     bigFloatArray[index % ARRAYSIZE] = data;   
}


/* 
 * Returns one float number read from index location
 */
float readBigArray(unsigned long int index)
{
  *BSEL = index / ARRAYSIZE;    // Set the Bank Select Register  
   return ( bigFloatArray[index % ARRAYSIZE]);
}
