//========================================================================
// SCIDYNE Corporation   649 School Street   Pembroke, MA 02359-3649 USA
// Tel: (781) 293-3059   Fax: (781) 293-4034   URL: http://www.scidyne.com 
//========================================================================
// NOTICE:
//    This work is licensed under a Creative Commons Attribution-ShareAlike 3.0 Unported License. 
//    http://creativecommons.org/licenses/by-sa/3.0/us/
//
//    and is based on the original work of Andy Brown:  
//    http://andybrown.me.uk/wk/2011/08/28/512kb-sram-expansion-for-the-arduino-mega-software
//========================================================================
// Project  :  XMEM+                                        
// File     :  xmem_plus_ex2.ino                                             
// Revision :  1.00                                               
// Date     :  12-11-15 
// Author:  :  Mark Durgin
// Target   :  Arduino Mega 2560, Arduino Mega ADK
// Compiler :  Arduino IDE 1.6.6                                                
//========================================================================
// Description and usage: 
//  This example outlines bank switching on the XMEM+. Three banks will
//  be used raising total SRAM to 104K (8K Internal + 96K using Banks 0,1,2).
//
//  The external memory feature of the MEGA is enabled and the Arduino heap is
//  relocated to use XMEM+ SRAM. Areas are created in each of three banks to
//  demostrate how to allocate dynamic memory and switch banks. Values are
//  then stored in those areas, read and and displayed on the IDE serial monitor.
//  
//========================================================================
// Issues, Concerns and "THINGS TO DO":
//
//========================================================================
// History:
//  12-11-15:
//    - Sketch first written
//========================================================================

/*******************************************************************************************
 * Important program parameters and constants
 *******************************************************************************************/
 #define XMEM_START            ( (void *) 0x2200 )   // Start of XMEM+ External SRAM memory space
 #define XMEM_END              ( (void *) 0xA1FF )   // End of XMEM+ External SRAM memory space
 #define MAX_BANKS_USED        3                     // Number of XMEM+ memory banks used. Adjust as needed; 1 to 16
 #define BSEL                  ( (char *) 0xFF01 )   // Pointer to XMEM+ Bank Select Register

 const unsigned char bd [] = {5,4,3,2,1,0};   // An array of arbitrary values to be used with bank1


/*******************************************************************************************
  Prototypes for the bank switching management functions
 *******************************************************************************************/
 void begin( bool heapInXmem_ );                          // Setup External Memeory, bank switching, and Expansion Bus  
 void setMemoryBank( uint8_t bank_, bool switchHeap_);    // Sets the active bank
 void saveHeap( uint8_t bank_ );                          // Saves a banks heap variables
 void restoreHeap( uint8_t bank_ );                       // Restores a banks heap variables


/*******************************************************************************************
   References to external system heap variables
 *******************************************************************************************/
 extern void * __heap_start;   // Start of heap from linker
 extern void * __heap_end;     // End of heap from linker
 extern void * __flp;          // Free-List pointer
 extern void * __brkval;       // Highest address used in heap


 /*******************************************************************************************
    Global variables, created in the .bss area of Arduino memory
  *******************************************************************************************/ 
 // An array of structures that hold heap variables associated with each bank
 struct heapState {
   char *__malloc_heap_start;          // Stores starting address of heap space
   char *__malloc_heap_end;            // Stores ending address of heap space
   void *__brkval;                     // Stores highest address used in heap
   void *__flp;                        // Stores "Free-List" Data Structure  pointer
 } bankHeapStates[ MAX_BANKS_USED ];   // Create an array of this structure, one structure for each bank

 
 
/********************************************************************************************
   Program initialization
 ********************************************************************************************/
void setup()
{
  begin( true );   // Setup for using External Memeory, bank switching, and Expansion Bus        
 
  // Setup communications to IDE serial monitor
  Serial.begin(9600);  

  // ...
  // Any other user initialization code goes here
  // ...

}



/********************************************************************************************
   This is the main loop
 ********************************************************************************************/
void loop()
{

 // Local variables
 char buf[80];     // Serial Monitor display buffer 
 int i;            // General purpose integer

 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
    For the purpose of this example show how to allocate dynamic memory areas and create
    pointers in differnt banks. Each pointer holds an areas starting address that was
    returned after calling dynamic memory functions such as malloc(), calloc()
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 // Bank-0 allocations and pointers
 setMemoryBank( 0, true);  // Switch  to bank-0
 char *  comBuf_bank0  = (char *) malloc( 10000 * sizeof(char));               // Request a 10,000 char area
 int *   xtraBuf_bank0 = (int *) malloc( 6500 * sizeof(int));                  // Request a 6,500 int (13,000 byte) area, Note: each int = 2 bytes
 byte *  tinyBuf_bank0 = (byte *) calloc( 48 * sizeof(byte), sizeof(byte));    // Request just 48 bytes. Use calloc() to have the area zeroed too
         
 // Bank-1 allocations and pointers
 setMemoryBank( 1, true);  // Switch  to bank-1
 unsigned char *  SDBuf_bank1 = (unsigned char *) malloc( 25000 * sizeof(unsigned char));   // Request a 25,000 unsigned char area 
 
 // Bank-2 allocations and pointers
 setMemoryBank( 2, true); // Switch  to bank-2
 char *  dispBuf_bank2 = (char *) malloc( 18000 * sizeof(char));   // Request an 18,000 char area
 
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
    With the dynamic memory areas and pointers now created show a few ways to use them 
    Note: bank access order doesn't matter as long as setMemoryBank() is called first
    to properly set the active bank and its related heap parameters
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
  // Load locations in bank-2 ... using sprintf() function  
 setMemoryBank( 2, true);   // Switch  to bank-2
 sprintf(dispBuf_bank2, "Hello from bank-2");  // Place a string in dispBuf_bank2 

 // Load locations in bank-0 ... like an integer array
 setMemoryBank( 0, true);   // Switch  to bank-0
 xtraBuf_bank0[0] = 0x0040;     // Load first location with a value
 xtraBuf_bank0[1] = 0xF506;     // Load second location with a value
 xtraBuf_bank0[5678] = 0x1234;  // Load a much higher location with a value

 // Load locations in bank-1 ... using memcpy() function
 setMemoryBank( 1, true);   // Switch  to bank-1
 memcpy(SDBuf_bank1, bd, sizeof(bd));  // Copy data from array

 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * 
    Display the various heap allocated addresses and the previously loaded values
    on the IDE Serial Monitor 
  * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 Serial.println(F("*** Addresses allocated on heap in different XMEM+ banks ***"));
 sprintf(buf,"comBuf_bank0   start address  : %04X", comBuf_bank0);
 Serial.println(buf);
 sprintf(buf,"xtraBuf_bank0  start address  : %04X", xtraBuf_bank0);
 Serial.println(buf);
 sprintf(buf,"tinyBuf_bank0  start address  : %04X", tinyBuf_bank0);
 Serial.println(buf);
 sprintf(buf,"SDBuf_bank1    start address  : %04X", SDBuf_bank1);
 Serial.println(buf);
 sprintf(buf,"dispBuf_bank2  start address  : %04X", dispBuf_bank2);
 Serial.println(buf);
 Serial.println();
 
 // Display the previously loaded values
 Serial.println(F("*** Information read from different banks ***"));
 
 Serial.println(F("Bank-0:"));
 setMemoryBank( 0, true);    // Switch  to bank-0
 sprintf(buf,"xtraBuf_bank0[0]    : %04X", xtraBuf_bank0[0]);
 Serial.println(buf);
 sprintf(buf,"xtraBuf_bank0[1]    : %04X", xtraBuf_bank0[1]);
 Serial.println(buf);
 sprintf(buf,"xtraBuf_bank0[5678] : %04X", xtraBuf_bank0[5678]);
 Serial.println(buf); Serial.println();
 
 setMemoryBank( 1, true);    // Switch  to bank-1
 Serial.println(F("Bank-1:"));  
 for (i = 0; i < sizeof(bd); i++)
 {
    sprintf(buf,"SDBuf_bank1[%d]: %d", i, SDBuf_bank1[i]);
    Serial.println(buf);
 }
 Serial.println();
 
 Serial.println(F("Bank-2:"));
 setMemoryBank( 2, true);  // Switch  to bank-2
 sprintf(buf,"dispBuf_bank2: %s", dispBuf_bank2);
 Serial.println(buf);

  // ...
  // Any other "run-once" user code goes here
  // ...

  
 while( true )    // Stay in this while loop forever
 {
    // ...
    // Any other user code goes here
    // ...  
 }
  
}



/*******************************************************************************************
 * Initial setup. You must call this once
 * heapInXmem_
 *  If = 1 (true) then the Arduino heap starting and ending addresses are set to use the XMEM+ SRAM.
 *  If = 0 (false) the only the external memory interface is activated.  The heap address ARE  NOT 
 *  reassigned to XMEM+ SRAM.
 *******************************************************************************************/
 void begin( bool heapInXmem_ ) {

   uint8_t bank;    // Temp value used to identify current bank

   // Activate external memory
   XMCRB = 0x00;             // All of all PORTC pins act as upper address lines, A[15:8], disable Bus keeper function 
   XMCRA = 0x80;             // Set SRE bit, Enable xmem+, No wait states. If Wait-states are needed change this value accordingly

   // initialize the heap states
   if( heapInXmem_ ) {
        __malloc_heap_end   =  static_cast <char *> (XMEM_END);
        __malloc_heap_start =  static_cast <char *> (XMEM_START);
    }

   for( bank = 0; bank < MAX_BANKS_USED; bank++ )
      saveHeap( bank );

   // set the current bank to bank zero
   setMemoryBank( 0, false );
 }

  

  /*******************************************************************************************
  * Set the memory bank - Call this to switch between the XMEM+ memory banks.
  * 
  * bank_
  *  Sets which bank to make active, valid range is 0 - 15  (0x00 - 0x0F)
  *
  * switchHeap_
  *  If 1 (true) the currently active bank heap parameters are saved, the new bank_ is
  *   made active, and it's heap parameters are loaded as the Arduino system heap parameters.
  *  If 0 (false) the bank_ is switched but no changes to the Arduino heap parameters are made.
  *
  *******************************************************************************************/
  void setMemoryBank( uint8_t  bank_, bool switchHeap_ ) {

    // check if requested bank is already active
    if(bank_== *BSEL )
      return;

    if( switchHeap_ )
    {  // Save heap state for the current bank if requested
      saveHeap( *BSEL );      // Save the currently active heap parameters
      *BSEL = bank_;          // switch in the new bank
      restoreHeap( bank_ );   // Restore the heap parameters for the new bank
    }
    else
    { // Just switch the bank, do not manage the heap parameters
     *BSEL = bank_;          // switch in the new bank          
    }
        
  }



  /*******************************************************************************************
   * Save the heap variables
   *******************************************************************************************/
  void saveHeap( uint8_t   bank_ ) {
    bankHeapStates[ bank_ ].__malloc_heap_start  = __malloc_heap_start;
    bankHeapStates[ bank_ ].__malloc_heap_end    = __malloc_heap_end;
    bankHeapStates[ bank_ ].__brkval             = __brkval;
    bankHeapStates[ bank_ ].__flp                = __flp;
  } 



  /*******************************************************************************************
   * Restore the heap variables
   *******************************************************************************************/
  void restoreHeap( uint8_t bank_ ) {
    __malloc_heap_start       = bankHeapStates[ bank_ ].__malloc_heap_start;
    __malloc_heap_end         = bankHeapStates[ bank_ ].__malloc_heap_end;
    __brkval                  = bankHeapStates[ bank_ ].__brkval;
    __flp                     = bankHeapStates[ bank_ ].__flp;
  }




