//========================================================================
// SCIDYNE Corporation   649 School Street   Pembroke, MA 02359-3649 USA
// Tel: (781) 293-3059   Fax: (781) 293-4034   URL: http://www.scidyne.com 
//========================================================================
// NOTICE:
//    This work is licensed under a Creative Commons Attribution-ShareAlike 3.0 Unported License. 
//    http://creativecommons.org/licenses/by-sa/3.0/us/
//
//    and is based on the original work of Andy Brown:  
//    http://andybrown.me.uk/wk/2011/08/28/512kb-sram-expansion-for-the-arduino-mega-software/
//========================================================================
// Project  :  XMEM+                                        
// File     :  xmem_plus.ino                                             
// Revision :  1.01                                               
// Date     :  12-03-15 
// Author:  :  Mark Durgin
// Target   :  Arduino Mega 2560, Arduino Mega ADK
// Compiler :  Arduino                                                 
//========================================================================
// Description and usage: 
//    This program merges the existing test_xmem software with the xmem works of andy brown.
//    Note: The xmem_plus.h and xmem_plus.ino files located in this directory are
//          required and appear as additional tabs in within the Arduino IDE.
//
//    There exist a few important differences which the sofware has been modified to accommodate for:
//    - The XMEM+ has 16 banks of 32KB per bank (0x2200 - 0xA1FF), NOT 56K (0x2200 - 0xFFFF)
//    - The XMEM+ has On-Board logic to control bank switching and so doesn't use the PORTD.7, PORTL.6, or PORTL.7 bits
//      as bank selector pins (address lines A16..A18) these are on pins 38,42,43 (PD7,PL7,PL6)
//
//  Web References:
//	http://www.nongnu.org/avr-libc/user-manual/malloc.html
//
//      http://jeelabs.org/2011/05/22/atmega-memory-use/
//
//      http://andybrown.me.uk/wk/2011/08/28/512kb-sram-expansion-for-the-arduino-mega-software/
//========================================================================
// Issues, Concerns and "THINGS TO DO":
//
//========================================================================
//    History:
//     12-04-15:
//     - Cleaned-up comments
//     - Revised setMemoryBank to support bank switching only, without heap management
//     - Updated SelfTestResults by adding missing false parameter when calling setMemoryBank( bank , false ); 
//     07-28-15:
//     - Changed the struct heapState to be a definition and an array at the same time.
//     07-25-15: 
//     - Added comments
//     - Commented out  #include <avr/io.h> 
//     - Changed how banks are controlled. The XMEM+ On-Board logic eliminates the requirement for the PD7, PD6, or PL7 bits
//     - Eliminated the variable  currentBank  becasue the XMEM+ BSEL register, being writable and readable, can be used for this purpose
//     - Changed hard-coded references to the number of banks from 8 to use constant MAX_BANKS  (which is 16 for the XMEM+)
//     - Added code to cause an error to test the test routine.  **** IMPORTANT: Comment Out when not using ****
//     - Changed hard-coded references to use constant MIN_WR_RD_TESTVALUE and MAX_WR_RD_TESTVALUE instead of "magic" numbers 
//========================================================================

//#include <avr/io.h>
#include "xmem_plus.h"

namespace xmem {

	/*******************************************************************************************
	 * State variables used by the heap
	 *******************************************************************************************/
	struct heapState {
	  char *__malloc_heap_start;          // Stores starting address of heap space
	  char *__malloc_heap_end;            // Stores ending address of heap space
	  void *__brkval;                     // Stores highest address used in heap
	  void *__flp;                        // Stores "Free-List" Data Structure  pointer
	} bankHeapStates[ MAX_BANKS_USED ];   // Create an array of this structure, one structure for each bank


	/*******************************************************************************************
	 * Initial setup. You must call this once
   * heapInXmem_
   *  If = 1 (true) then the Arduino heap starting and ending addresses are set to use the XMEM+ SRAM.
   *  If = 0 (false) the only the external memory interface is activated.  The heap address ARE  NOT 
   *  reassigned to XMEM+ SRAM.
	 *******************************************************************************************/
	void begin( bool heapInXmem_ ) {

	  uint8_t   bank;    // Temp value used to identify current bank

	  // set up the xmem registers
          // Activate external memory
          XMCRB  =  0x00;             // All of all PORTC pins act as upper address lines, A[15:8], disable Bus keeper function 
          XMCRA  =  0x80;             // Set SRE bit, Enable xmem+, No wait states. If Wait-states are needed change this value accordingly

	  // initialise the heap states
	  if( heapInXmem_ ) {
	    __malloc_heap_end   =  static_cast <char *> (XMEM_END);
	    __malloc_heap_start =  static_cast <char *> (XMEM_START);
	  }

	  for(bank = 0; bank < MAX_BANKS_USED; bank++)
	    saveHeap( bank );

	  // set the current bank to zero
	  setMemoryBank(0, false);
	}

	

  /*******************************************************************************************
	* Set the memory bank - Call this to switch between the XMEM+ memory banks.
  * 
  * bank_
  *  Sets which bank to make active, valid range is 0 - 15  (0x00 - 0x0F)
  *
  * switchHeap_
  *  If 1 (true) the currently active bank heap parameters are saved, the new bank_ is
  *   made active, and it's heap paramters are loaded as the Arduino system heap parameters.
  *  If 0 (false) the bank_ is switched but no changes to the Arduino heap parameters are made.
  *
	*******************************************************************************************/
	void setMemoryBank(uint8_t   bank_,   bool switchHeap_) {

  	// check if requested bank is already active
	  if(bank_== *BSEL )
	    return;

	  if( switchHeap_ )
	  // Save heap state for the current bank if requested
	  {  
      saveHeap( *BSEL );      // Save the currently active heap parameters
      *BSEL = bank_;          // switch in the new bank
      restoreHeap( bank_ );   // Restore the heap parameters for this bank
	  }
    else
    // Just switch the bank, do not manage the heap parameters
    {
     *BSEL = bank_;          // switch in the new bank          
    }
        
  }




	/*******************************************************************************************
	 * Save the heap variables
	 *******************************************************************************************/
	void saveHeap( uint8_t   bank_ ) {
	  bankHeapStates[ bank_ ].__malloc_heap_start  = __malloc_heap_start;
	  bankHeapStates[ bank_ ].__malloc_heap_end    = __malloc_heap_end;
	  bankHeapStates[ bank_ ].__brkval             = __brkval;
	  bankHeapStates[ bank_ ].__flp                = __flp;
	} 



	/*******************************************************************************************
	 * Restore the heap variables
	 *******************************************************************************************/
	void restoreHeap( uint8_t bank_ ) {
	  __malloc_heap_start       = bankHeapStates[ bank_ ].__malloc_heap_start;
	  __malloc_heap_end         = bankHeapStates[ bank_ ].__malloc_heap_end;
	  __brkval                  = bankHeapStates[ bank_ ].__brkval;
	  __flp                     = bankHeapStates[ bank_ ].__flp;
	}



	/*******************************************************************************************
	 * Self test the memory. This will destroy the entire contents of all
	 * memory banks so don't use it except in a test scenario.
	 *******************************************************************************************/
	SelfTestResults selfTest() {

	  volatile          uint8_t   *ptr;
	  uint8_t           bank, writeValue, readValue;
	  SelfTestResults   results;

	  // write an ascending sequence of MIN_WR_RD_TESTVALUE .... MAX_WR_RD_TESTVALUE running through all memory banks
          
          writeValue = MIN_WR_RD_TESTVALUE;
		
          for( bank = 0;    bank  <  MAX_BANKS_USED;    bank++ ) {
            setMemoryBank( bank , false );

	    for( ptr = reinterpret_cast <uint8_t *> (XMEM_END);  ptr >=  reinterpret_cast <uint8_t *> (XMEM_START);  ptr-- ) {
	      *ptr = writeValue;

	      if( writeValue++ == MAX_WR_RD_TESTVALUE )
	        writeValue = MIN_WR_RD_TESTVALUE;
			
                // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                // Uncomment the next two lines of code to force a error when verifing
                // this test software. As shown here an error will be forced in bank #3
                // at address 0x6543. The value 165 (0xA5) will be stored which over-writes
                // the previously written value done in the above loop.
                // The bank, address, and value can be changed as needed.
                // > > > Important: Comment out the two lines of code if not using < < <
                // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
                // if ( ( bank == 3 )  && ( ptr == (uint8_t *) 0x6543 ) )
                //    *ptr = 165;  
             }
          }

 	
          // verify the writes
	  writeValue = MIN_WR_RD_TESTVALUE;
	
          for( bank  = 0;   bank  <  MAX_BANKS_USED;   bank++ ) {

	    setMemoryBank( bank, false );

	    for( ptr = reinterpret_cast <uint8_t *> (XMEM_END); ptr >= reinterpret_cast <uint8_t *> (XMEM_START); ptr-- ) {

	      readValue = *ptr;

	      if( readValue != writeValue ) {
	        results.succeeded     = false;
		      results.failedAddress = ptr;
		      results.failedBank    = bank;
		      return results;
	      }

	      if( writeValue++ == MAX_WR_RD_TESTVALUE )
	        writeValue = MIN_WR_RD_TESTVALUE;
	      }
	    }

	    results.succeeded = true;
	    return ( results );
	  }
        }

